import asyncio
import json
from typing import Optional

from aiohttp import ClientSession
from aiohttp_sse_client import client as sse_client
from pytonconnect.crypto import SessionCrypto

from pytonconnect.provider import BridgeProvider as BaseBridgeProvider
from pytonconnect.provider import BridgeGateway as BaseBridgeGateway
from pytonconnect.provider._bridge_session import BridgeSession  # noqa
from pytonconnect.storage import IStorage
from pytonconnect.logger import _LOGGER  # noqa


class BridgeGateway(BaseBridgeGateway):

    def __init__(
            self,
            storage: IStorage,
            bridge_url: str,
            session_id: str,
            listener,
            errors_listener,
            tonapi_token: Optional[str] = None,
    ) -> None:
        super().__init__(storage, bridge_url, session_id, listener, errors_listener)
        self.tonapi_token = tonapi_token

    async def send(self, request: str, receiver_public_key: str, topic: str, ttl: int = None):
        bridge_base = self._bridge_url.rstrip('/')
        bridge_url = f'{bridge_base}/{self.POST_PATH}?client_id={self._session_id}'
        bridge_url += f'&to={receiver_public_key}'
        bridge_url += f'&ttl={ttl if ttl else self.DEFAULT_TTL}'
        bridge_url += f'&topic={topic}'
        headers = {'Content-type': 'text/plain;charset=UTF-8'}

        if "tonapi" in bridge_base and self.tonapi_token:
            headers["Authorization"] = f"Bearer {self.tonapi_token}"

        async with ClientSession(headers=headers) as session:
            async with session.post(bridge_url, data=request):
                pass

    async def register_session(self) -> bool:
        if self._is_closed:
            return False

        bridge_base = self._bridge_url.rstrip('/')
        bridge_url = f'{bridge_base}/{self.SSE_PATH}?client_id={self._session_id}'

        last_event_id = await self._storage.get_item(IStorage.KEY_LAST_EVENT_ID)
        if last_event_id:
            bridge_url += f'&last_event_id={last_event_id}'
        _LOGGER.debug(f'Bridge url -> {bridge_url}')

        if self._handle_listen is not None:
            self._handle_listen.cancel()

        loop = asyncio.get_running_loop()
        resolve = loop.create_future()

        session = ClientSession()
        if "tonapi" in bridge_base and self.tonapi_token:
            headers = {"Authorization": f"Bearer {self.tonapi_token}"}
            session.headers.update(headers)

        self._event_source = sse_client.EventSource(
            bridge_url,
            timeout=-1,
            session=session,
            on_error=self._errors_handler
        )
        self._handle_listen = asyncio.create_task(self.listen_event_source(resolve))

        return await resolve


class BridgeProvider(BaseBridgeProvider):
    _gateway: BridgeGateway

    def __init__(self, storage: IStorage, wallet: dict, redirect_url: str = None, tonapi_token: str = None) -> None:
        self.redirect_url = redirect_url
        self.tonapi_token = tonapi_token
        super().__init__(storage, wallet)

    async def connect(self, request: dict):
        self._close_gateways()
        session_crypto = SessionCrypto()

        bridge_url = ''
        universal_url = BridgeProvider.STANDART_UNIVERSAL_URL

        if isinstance(self._wallet, dict):
            bridge_url = self._wallet['bridge_url']
            if 'universal_url' in self._wallet:
                universal_url = self._wallet['universal_url']

            self._gateway = BridgeGateway(self._storage, bridge_url, session_crypto.session_id, self._gateway_listener,
                                          self._gateway_errors_listener, tonapi_token=self.tonapi_token)

            await self._gateway.register_session()

        self._session.session_crypto = session_crypto
        self._session.bridge_url = bridge_url

        return self._generate_universal_url(universal_url, request)

    async def restore_connection(self):
        self._close_gateways()

        connection = await self._storage.get_item(IStorage.KEY_CONNECTION)
        if not connection:
            return False
        connection = json.loads(connection)

        if 'session' not in connection:
            return False
        self._session = BridgeSession(connection['session'])

        self._gateway = BridgeGateway(self._storage, self._session.bridge_url, self._session.session_crypto.session_id,
                                      self._gateway_listener, self._gateway_errors_listener,
                                      tonapi_token=self.tonapi_token)

        await self._gateway.register_session()

        for listener in self._listeners:
            listener(connection['connect_event'])

        return True

    async def _gateway_listener(self, bridge_incoming_message):
        wallet_message = json.loads(
            self._session.session_crypto.decrypt(bridge_incoming_message['message'], bridge_incoming_message['from'])
        )

        _LOGGER.debug(f'Wallet message received: {wallet_message}')

        if 'event' not in wallet_message:
            if 'id' in wallet_message:
                id_ = wallet_message['id']
                if id_ not in self._pending_requests:
                    _LOGGER.debug(f"Response id {id_} doesn't match any request's id")
                    return

                self._pending_requests[id_].set_result(wallet_message)
                del self._pending_requests[id_]
            return

        if 'id' in wallet_message:
            id_ = int(wallet_message['id'])
            connection = json.loads(await self._storage.get_item(IStorage.KEY_CONNECTION, '{}'))
            last_id = connection['last_wallet_event_id'] if 'last_wallet_event_id' in connection else 0

            if last_id and id_ <= last_id:
                id_ = id_ - 1 if id_ == last_id else id_ - 2
                _LOGGER.error(
                    f'Received event id (={id_}) must be greater than stored last wallet event id (={last_id})')

            if 'event' in wallet_message and wallet_message['event'] != 'connect':
                connection['last_wallet_event_id'] = id_
                await self._storage.set_item(IStorage.KEY_CONNECTION, json.dumps(connection))

        # self.listeners might be modified in the event handler
        listeners = self._listeners.copy()

        if wallet_message['event'] == 'connect':
            await self._update_session(wallet_message, bridge_incoming_message['from'])

        elif wallet_message['event'] == 'disconnect':
            await self._remove_session()

        for listener in listeners:
            listener(wallet_message)

    def _generate_tg_universal_url(self, universal_url: str, request: dict):
        universal_url = universal_url.replace("attach=wallet", "startattach=")
        link = self._generate_regular_universal_url('about:blank', request)
        link_params = link.split('?')[1]

        link_params += f"&ret={self.redirect_url}" if self.redirect_url else "&ret=back"

        start_attach = (
                'tonconnect-' + link_params
                .replace('.', '%2E')
                .replace('-', '%2D')
                .replace('_', '%5F')
                .replace('&', '-')
                .replace('=', '__')
                .replace('%', '--')
                .replace('+', '')
                .replace(':', '--3A')
                .replace('/', '--2F')
        )
        return universal_url + start_attach



from abc import ABCMeta, abstractmethod
from typing import List, Dict

from aiogram.utils.keyboard import InlineKeyboardMarkup as Markup, InlineKeyboardBuilder
from aiogram.utils.keyboard import InlineKeyboardButton as Button

from ..tonconnect.models import AppWallet


class InlineKeyboardBase(metaclass=ABCMeta):
    """
    Abstract base class for handling inline keyboards with buttons.

    :param language_code: The language code for the button texts.
    """

    @property
    @abstractmethod
    def texts_buttons(self) -> Dict[str, Dict[str, str]]:
        """
        Property to retrieve a dictionary of button texts for different languages.

        :return: Dictionary containing button texts for each language.
        """
        raise NotImplementedError

    def __init__(self, language_code: str) -> None:
        """
        Initialize the InlineKeyboardBase with a specified language code.

        If the provided language code is not supported, the default language is set to "en" (English).

        :param language_code: The language code for the button texts.
        """
        if language_code not in self.texts_buttons.keys():
            language_code = "en"
        self.language_code = language_code

    @abstractmethod
    def connect_wallet(
            self,
            wallets: List[AppWallet],
            selected_wallet: AppWallet,
            universal_url: str,
            wallet_name: str,
            width: int = 2,
    ) -> Markup:
        """
        Create an inline keyboard for connecting a wallet.

        :param wallets: List of available app wallets.
        :param selected_wallet: The selected app wallet.
        :param universal_url: Universal URL for connecting the wallet.
        :param wallet_name: Name of the wallet.
        :param width: Number of wallet buttons per row.
        :return: Inline keyboard markup.
        """
        raise NotImplementedError

    @abstractmethod
    def connect_wallet_proof_wrong(self) -> Markup:
        """
        Create an inline keyboard for proof of wrong connection.

        :return: Inline keyboard markup.
        """
        raise NotImplementedError

    @abstractmethod
    def send_transaction(self, wallet_name: str, url: str) -> Markup:
        """
        Create an inline keyboard for sending a transaction.

        :param wallet_name: Name of the wallet.
        :param url: URL for opening the wallet.
        :return: Inline keyboard markup.
        """
        raise NotImplementedError

    @abstractmethod
    def send_transaction_timeout(self) -> Markup:
        """
        Create an inline keyboard for a transaction timeout.

        :return: Inline keyboard markup.
        """
        raise NotImplementedError

    @abstractmethod
    def send_transaction_rejected(self) -> Markup:
        """
        Create an inline keyboard for a rejected transaction.

        :return: Inline keyboard markup.
        """
        raise NotImplementedError

    def _get_button(
            self,
            code: str,
            url: str = None,
            **kwargs,
    ) -> Button:
        """
        Helper method to create an inline keyboard button.

        :param code: Code identifying the specific button.
        :param url: URL associated with the button.
        :param kwargs: Additional arguments for formatting button text.
        :return: Inline keyboard button.
        """
        text = self.texts_buttons[self.language_code][code].format_map(kwargs)
        if not url:
            return Button(text=text, callback_data=code)
        return Button(text=text, url=url)


class InlineKeyboard(InlineKeyboardBase):

    @property
    def texts_buttons(self) -> Dict[str, Dict[str, str]]:
        return {
            "ru": {
                "back": "‹ Назад",
                "retry": "↻ Повторить",
                "connect_wallet": "Подключить {wallet_name}",
                "open_wallet": "Перейти в {wallet_name}",
            },
            "en": {
                "back": "‹ Back",
                "retry": "↻ Retry",
                "connect_wallet": "Connect {wallet_name}",
                "open_wallet": "Go to {wallet_name}",
            },
        }

    def connect_wallet(
            self,
            wallets: List[AppWallet],
            selected_wallet: AppWallet,
            universal_url: str,
            wallet_name: str,
            width: int = 2,
    ) -> Markup:
        generated_buttons = [
            *[
                Button(
                    text=f"• {wallet.name} •" if wallet.app_name == selected_wallet.app_name else wallet.name,
                    callback_data=f"app_wallet:{wallet.app_name}",
                ) for wallet in wallets
            ]
        ]
        builder = InlineKeyboardBuilder()
        builder.row(self._get_button("connect_wallet", universal_url, wallet_name=wallet_name))
        builder.row(*generated_buttons, width=width)
        builder.row(self._get_button("back"))
        return builder.as_markup()

    def connect_wallet_proof_wrong(self) -> Markup:
        inline_keyboard = [
            [self._get_button("back"),
             self._get_button("retry")],
        ]
        return Markup(inline_keyboard=inline_keyboard)

    def send_transaction(self, wallet_name: str, url: str) -> Markup:
        inline_keyboard = [
            [self._get_button("open_wallet", url=url, wallet_name=wallet_name)],
            [self._get_button("back")],
        ]
        return Markup(inline_keyboard=inline_keyboard)

    def send_transaction_timeout(self) -> Markup:
        inline_keyboard = [
            [self._get_button("back"),
             self._get_button("retry")],
        ]
        return Markup(inline_keyboard=inline_keyboard)

    def send_transaction_rejected(self) -> Markup:
        inline_keyboard = [
            [self._get_button("back"),
             self._get_button("retry")],
        ]
        return Markup(inline_keyboard=inline_keyboard)



from typing import List, Optional

from pytonconnect import TonConnect as BaseTonConnect
from pytonconnect.logger import _LOGGER  # noqa

from . import SessionStorage
from .provider import BridgeProvider
from .wallet.manager import WalletManager
from .models import AppWallet


class AiogramTonConnect(BaseTonConnect):
    """
    AiogramTonConnect class that extends the base AiogramTonConnect functionality.

    :param storage: An instance of SessionStorage for session data storage.
    :param manifest_url: URL to the manifest file.
    :param redirect_url: URL to the redirect after connecting.
    :param exclude_wallets: Optional list of wallet names to exclude.
    :param args: Additional positional arguments.
    :param kwargs: Additional keyword arguments.
    """

    def __init__(
            self,
            storage: SessionStorage,
            manifest_url: str,
            redirect_url: str = None,
            exclude_wallets: Optional[List[str]] = None,
            tonapi_token: Optional[str] = None,
            *args,
            **kwargs,
    ) -> None:
        kwargs["storage"] = storage
        kwargs["manifest_url"] = manifest_url
        super().__init__(*args, **kwargs)
        self.redirect_url = redirect_url
        self.wallet_manager = WalletManager(exclude_wallets=exclude_wallets)
        self.tonapi_token = tonapi_token

    async def get_wallets(self=None) -> List[AppWallet]:
        """
        Get the list of AiogramTonConnect wallets.

        :return: List of AiogramTonConnect application wallet instances.
        """
        return await self.wallet_manager.get_wallets()

    def _create_provider(self, wallet: dict) -> BridgeProvider:
        provider = BridgeProvider(
            self._storage,
            wallet,
            redirect_url=self.redirect_url,
            tonapi_token=self.tonapi_token,
        )
        provider.listen(self._wallet_events_listener)
        self._provider = provider
        return provider

    async def disconnect(self):
        await super().disconnect()
        await self._storage.remove_item(SessionStorage.KEY_CONNECTION)
        await self._storage.remove_item(SessionStorage.KEY_LAST_EVENT_ID)



import pickle

from .base import ATCStorageBase
from ..models import ConnectWalletCallbacks, SendTransactionCallbacks


class ConnectWalletCallbackStorage:

    def __init__(
            self,
            storage: ATCStorageBase,
            user_id: int,
            collection: str = "ConnectWalletCallbacks",
    ) -> None:
        self.storage = storage
        self.user_id = user_id
        self.collection = collection

    def _get_key(self) -> str:
        """
        Generate a unique key for the session storage.

        :return: Unique key combining the collection and user_id.
        """
        return f"{self.collection}:{self.user_id}"

    async def get(self) -> ConnectWalletCallbacks:
        value = await self.storage.get_item(self._get_key())
        return ConnectWalletCallbacks(**pickle.loads(value)) if value else None

    async def add(self, connect_wallet_callbacks: ConnectWalletCallbacks) -> None:
        serialized_value = pickle.dumps(connect_wallet_callbacks.model_dump())
        await self.storage.set_item(self._get_key(), serialized_value)

    async def remove(self) -> None:
        await self.storage.remove_item(self._get_key())


class SendTransactionCallbackStorage:

    def __init__(
            self,
            storage: ATCStorageBase,
            user_id: int,
            collection: str = "SendTransactionsCallbacks",
    ) -> None:
        self.storage = storage
        self.user_id = user_id
        self.collection = collection

    def _get_key(self) -> str:
        """
        Generate a unique key for the session storage.

        :return: Unique key combining the collection and user_id.
        """
        return f"{self.collection}:{self.user_id}"

    async def get(self) -> SendTransactionCallbacks:
        value = await self.storage.get_item(self._get_key())
        return SendTransactionCallbacks(**pickle.loads(value)) if value else None

    async def add(self, send_transaction_callbacks: SendTransactionCallbacks) -> None:
        serialized_value = pickle.dumps(send_transaction_callbacks.model_dump())
        await self.storage.set_item(self._get_key(), serialized_value)

    async def remove(self) -> None:
        await self.storage.remove_item(self._get_key())



import base64
from abc import ABCMeta, abstractmethod

from .generator import generate_qrcode

__all__ = [
    "QRImageProviderBase",
    "QRUrlProviderBase",

    "QRImageProvider",
    "QRUrlProvider",
]


class QRImageProviderBase(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    async def create_connect_wallet_image(cls, universal_url: str, *args, **kwargs) -> bytes:
        """
        Create a QR code image for connecting a wallet.

        :param universal_url: Universal URL for connecting the wallet.
        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.
        :return: Bytes representing the QR code image.
        """
        raise NotImplementedError(
            "Implement the method to generate the QR code image "
            "in bytes from the universal_url data."
        )


class QRUrlProviderBase(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    async def create_connect_wallet_image_url(cls, universal_url: str, *args, **kwargs) -> str:
        """
        Create a URL for connecting a wallet and generate a QR code.

        :param universal_url: Universal URL for connecting the wallet.
        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.
        :return: URL for generating the QR code.
        """
        raise NotImplementedError(
            "Implement the method to generate the QR code and retry the URL string "
            "using your API from the universal_url data."
        )


class QRImageProvider(QRImageProviderBase):

    @classmethod
    async def create_connect_wallet_image(cls, universal_url: str, *args, **kwargs) -> bytes:
        return await generate_qrcode(
            universal_url,
            border=7,
            box_size=20,
            image_url=args[0],
            image_padding=20,
            image_round=50,
        )


class QRUrlProvider(QRUrlProviderBase):

    @classmethod
    async def create_connect_wallet_image_url(cls, universal_url: str, *args, **kwargs) -> str:
        return (
            f"https://qrcode.ness.su/create?"
            f"box_size=20&border=7&image_padding=20"
            f"&data={base64.b64encode(universal_url.encode()).decode()}"
            f"&image_url={base64.b64encode(args[0].encode()).decode()}"
        )
